@@ -35,6 +35,7 @@
 #include "netdev-offload-provider.h"
 #include "openvswitch/dynamic-string.h"
 #include "openvswitch/vlog.h"
+#include "ovs-atomic.h"
 #include "ovs-numa.h"
 #include "smap.h"
 #include "svec.h"
@@ -50,10 +51,11 @@ static char *vhost_sock_dir = NULL;   /* Location of vhost-user sockets */
 static bool vhost_iommu_enabled = false; /* Status of vHost IOMMU support */
 static bool vhost_postcopy_enabled = false; /* Status of vHost POSTCOPY
                                              * support. */
-static bool dpdk_initialized = false; /* Indicates successful initialization
-                                       * of DPDK. */
 static bool per_port_memory = false; /* Status of per port memory support */
 
+/* Indicates successful initialization of DPDK. */
+static atomic_bool dpdk_initialized = ATOMIC_VAR_INIT(false);
+
 static int
 process_vhost_flags(char *flag, const char *default_val, int size,
                     const struct smap *ovs_other_config,
@@ -130,74 +132,12 @@ construct_dpdk_options(const struct smap *ovs_other_config, struct svec *args)
     }
 }
 
-static int
-compare_numa_node_list(const void *a_, const void *b_)
-{
-    int a = *(const int *) a_;
-    int b = *(const int *) b_;
-
-    if (a < b) {
-        return -1;
-    }
-    if (a > b) {
-        return 1;
-    }
-    return 0;
-}
-
-static char *
-construct_dpdk_socket_mem(void)
-{
-    const char *def_value = "1024";
-    struct ds dpdk_socket_mem = DS_EMPTY_INITIALIZER;
-
-    /* Build a list of all numa nodes with at least one core. */
-    struct ovs_numa_dump *dump = ovs_numa_dump_n_cores_per_numa(1);
-    size_t n_numa_nodes = hmap_count(&dump->numas);
-    int *numa_node_list = xcalloc(n_numa_nodes, sizeof *numa_node_list);
-
-    const struct ovs_numa_info_numa *node;
-    int k = 0, last_node = 0;
-
-    FOR_EACH_NUMA_ON_DUMP(node, dump) {
-        if (k >= n_numa_nodes) {
-            break;
-        }
-        numa_node_list[k++] = node->numa_id;
-    }
-    qsort(numa_node_list, k, sizeof *numa_node_list, compare_numa_node_list);
-
-    for (int i = 0; i < n_numa_nodes; i++) {
-        while (numa_node_list[i] > last_node &&
-               numa_node_list[i] != OVS_NUMA_UNSPEC &&
-               numa_node_list[i] <= MAX_NUMA_NODES) {
-            if (last_node == 0) {
-                ds_put_format(&dpdk_socket_mem, "%s", "0");
-            } else {
-                ds_put_format(&dpdk_socket_mem, ",%s", "0");
-            }
-            last_node++;
-        }
-        if (numa_node_list[i] == 0) {
-            ds_put_format(&dpdk_socket_mem, "%s", def_value);
-        } else {
-            ds_put_format(&dpdk_socket_mem, ",%s", def_value);
-        }
-        last_node++;
-    }
-    free(numa_node_list);
-    ovs_numa_dump_destroy(dump);
-    return ds_cstr(&dpdk_socket_mem);
-}
-
 #define MAX_DPDK_EXCL_OPTS 10
 
 static void
 construct_dpdk_mutex_options(const struct smap *ovs_other_config,
                              struct svec *args)
 {
-    char *default_dpdk_socket_mem = construct_dpdk_socket_mem();
-
     struct dpdk_exclusive_options_map {
         const char *category;
         const char *ovs_dpdk_options[MAX_DPDK_EXCL_OPTS];
@@ -208,7 +148,7 @@ construct_dpdk_mutex_options(const struct smap *ovs_other_config,
         {"memory type",
          {"dpdk-alloc-mem", "dpdk-socket-mem", NULL,},
          {"-m",             "--socket-mem",    NULL,},
-         default_dpdk_socket_mem, 1
+         NULL, 0
         },
     };
 
@@ -217,7 +157,6 @@ construct_dpdk_mutex_options(const struct smap *ovs_other_config,
         int found_opts = 0, scan, found_pos = -1;
         const char *found_value;
         struct dpdk_exclusive_options_map *popt = &excl_opts[i];
-        bool using_default = false;
 
         for (scan = 0; scan < MAX_DPDK_EXCL_OPTS
                  && popt->ovs_dpdk_options[scan]; ++scan) {
@@ -234,7 +173,6 @@ construct_dpdk_mutex_options(const struct smap *ovs_other_config,
             if (popt->default_option) {
                 found_pos = popt->default_option;
                 found_value = popt->default_value;
-                using_default = true;
             } else {
                 continue;
             }
@@ -247,12 +185,6 @@ construct_dpdk_mutex_options(const struct smap *ovs_other_config,
         }
 
         if (!args_contains(args, popt->eal_dpdk_options[found_pos])) {
-            if (using_default) {
-                VLOG_INFO("Using default value for '%s'. OVS will no longer "
-                          "provide a default for this argument starting "
-                          "from 2.17 release. DPDK defaults will be used "
-                          "instead.", popt->eal_dpdk_options[found_pos]);
-            }
             svec_add(args, popt->eal_dpdk_options[found_pos]);
             svec_add(args, found_value);
         } else {
@@ -260,8 +192,6 @@ construct_dpdk_mutex_options(const struct smap *ovs_other_config,
                       "dpdk-extra config", popt->eal_dpdk_options[found_pos]);
         }
     }
-
-    free(default_dpdk_socket_mem);
 }
 
 static void
@@ -477,25 +407,12 @@ dpdk_init__(const struct smap *ovs_other_config)
     svec_add(&args, ovs_get_program_name());
     construct_dpdk_args(ovs_other_config, &args);
 
-    if (!args_contains(&args, "--legacy-mem")
-        && !args_contains(&args, "--socket-limit")) {
-        const char *arg;
-        size_t i;
-
-        SVEC_FOR_EACH (i, arg, &args) {
-            if (!strcmp(arg, "--socket-mem")) {
-                break;
-            }
-        }
-        if (i < args.n - 1) {
-            svec_add(&args, "--socket-limit");
-            svec_add(&args, args.names[i + 1]);
-            VLOG_INFO("Using default value for '--socket-limit'. OVS will no "
-                      "longer provide a default for this argument starting "
-                      "from 2.17 release. DPDK defaults will be used "
-                      "instead.");
-        }
+#ifdef DPDK_IN_MEMORY_SUPPORTED
+    if (!args_contains(&args, "--in-memory") &&
+            !args_contains(&args, "--legacy-mem")) {
+        svec_add(&args, "--in-memory");
     }
+#endif
 
     if (args_contains(&args, "-c") || args_contains(&args, "-l")) {
         auto_determine = false;
@@ -562,6 +479,12 @@ dpdk_init__(const struct smap *ovs_other_config)
         return false;
     }
 
+    if (!rte_mp_disable()) {
+        VLOG_EMER("Could not disable multiprocess, DPDK won't be available.");
+        rte_eal_cleanup();
+        return false;
+    }
+
     if (VLOG_IS_DBG_ENABLED()) {
         size_t size;
         char *response = NULL;
@@ -581,6 +504,8 @@ dpdk_init__(const struct smap *ovs_other_config)
         }
     }
 
+    unixctl_command_register("dpdk/lcore-list", "", 0, 0,
+                             dpdk_unixctl_mem_stream, rte_lcore_dump);
     unixctl_command_register("dpdk/log-list", "", 0, 0,
                              dpdk_unixctl_mem_stream, rte_log_dump);
     unixctl_command_register("dpdk/log-set", "{level | pattern:level}", 0,
@@ -593,7 +518,7 @@ dpdk_init__(const struct smap *ovs_other_config)
     RTE_PER_LCORE(_lcore_id) = NON_PMD_CORE_ID;
 
     /* Finally, register the dpdk classes */
-    netdev_dpdk_register();
+    netdev_dpdk_register(ovs_other_config);
     netdev_register_flow_api_provider(&netdev_offload_dpdk);
     return true;
 }
@@ -630,7 +555,7 @@ dpdk_init(const struct smap *ovs_other_config)
     } else {
         VLOG_INFO_ONCE("DPDK Disabled - Use other_config:dpdk-init to enable");
     }
-    dpdk_initialized = enabled;
+    atomic_store_relaxed(&dpdk_initialized, enabled);
 }
 
 const char *
@@ -660,15 +585,40 @@ dpdk_per_port_memory(void)
 bool
 dpdk_available(void)
 {
-    return dpdk_initialized;
+    bool initialized;
+
+    atomic_read_relaxed(&dpdk_initialized, &initialized);
+    return initialized;
 }
 
-void
-dpdk_set_lcore_id(unsigned cpu)
+bool
+dpdk_attach_thread(unsigned cpu)
 {
     /* NON_PMD_CORE_ID is reserved for use by non pmd threads. */
     ovs_assert(cpu != NON_PMD_CORE_ID);
-    RTE_PER_LCORE(_lcore_id) = cpu;
+
+    if (!dpdk_available()) {
+        return false;
+    }
+
+    if (rte_thread_register() < 0) {
+        VLOG_WARN("DPDK max threads count has been reached. "
+                  "PMD thread performance may be impacted.");
+        return false;
+    }
+
+    VLOG_INFO("PMD thread uses DPDK lcore %u.", rte_lcore_id());
+    return true;
+}
+
+void
+dpdk_detach_thread(void)
+{
+    unsigned int lcore_id;
+
+    lcore_id = rte_lcore_id();
+    rte_thread_unregister();
+    VLOG_INFO("PMD thread released DPDK lcore %u.", lcore_id);
 }
 
 void
@@ -677,63 +627,11 @@ print_dpdk_version(void)
     puts(rte_version());
 }
 
-/* Avoid calling rte_cpu_get_flag_enabled() excessively, by caching the
- * result of the call for each CPU flag in a static variable. To avoid
- * allocating large numbers of static variables, use a uint8 as a bitfield.
- * Note the macro must only return if the ISA check is done and available.
- */
-#define ISA_CHECK_DONE_BIT (1 << 0)
-#define ISA_AVAILABLE_BIT  (1 << 1)
-
-#define CHECK_CPU_FEATURE(feature, name_str, RTE_CPUFLAG)               \
-    do {                                                                \
-        if (strncmp(feature, name_str, strlen(name_str)) == 0) {        \
-            static uint8_t isa_check_##RTE_CPUFLAG;                     \
-            int check = isa_check_##RTE_CPUFLAG & ISA_CHECK_DONE_BIT;   \
-            if (OVS_UNLIKELY(!check)) {                                 \
-                int has_isa = rte_cpu_get_flag_enabled(RTE_CPUFLAG);    \
-                VLOG_DBG("CPU flag %s, available %s\n",                 \
-                         name_str, has_isa ? "yes" : "no");             \
-                isa_check_##RTE_CPUFLAG = ISA_CHECK_DONE_BIT;           \
-                if (has_isa) {                                          \
-                    isa_check_##RTE_CPUFLAG |= ISA_AVAILABLE_BIT;       \
-                }                                                       \
-            }                                                           \
-            if (isa_check_##RTE_CPUFLAG & ISA_AVAILABLE_BIT) {          \
-                return true;                                            \
-            } else {                                                    \
-                return false;                                           \
-            }                                                           \
-        }                                                               \
-    } while (0)
-
-bool
-dpdk_get_cpu_has_isa(const char *arch, const char *feature)
-{
-    /* Ensure Arch is x86_64. */
-    if (strncmp(arch, "x86_64", 6) != 0) {
-        return false;
-    }
-
-#if __x86_64__
-    /* CPU flags only defined for the architecture that support it. */
-    CHECK_CPU_FEATURE(feature, "avx512f", RTE_CPUFLAG_AVX512F);
-    CHECK_CPU_FEATURE(feature, "avx512bw", RTE_CPUFLAG_AVX512BW);
-    CHECK_CPU_FEATURE(feature, "avx512vbmi", RTE_CPUFLAG_AVX512VBMI);
-    CHECK_CPU_FEATURE(feature, "avx512vpopcntdq", RTE_CPUFLAG_AVX512VPOPCNTDQ);
-    CHECK_CPU_FEATURE(feature, "bmi2", RTE_CPUFLAG_BMI2);
-#endif
-
-    VLOG_WARN("Unknown CPU arch,feature: %s,%s. Returning not supported.\n",
-              arch, feature);
-    return false;
-}
-
 void
 dpdk_status(const struct ovsrec_open_vswitch *cfg)
 {
     if (cfg) {
-        ovsrec_open_vswitch_set_dpdk_initialized(cfg, dpdk_initialized);
+        ovsrec_open_vswitch_set_dpdk_initialized(cfg, dpdk_available());
         ovsrec_open_vswitch_set_dpdk_version(cfg, rte_version());
     }
 }
